message = """
Runnable modules contained here

  zgulde.nbslides     -- (WIP): turn jupyter notebooks into remark.js slides
  zgulde.gitlog       -- make git logs more machine readable
  zgulde.em           -- extract markdown documents from code comments
  zgulde.contact_info
""".strip()

print(message)
